import java.sql.Connection
import java.sql.ResultSet
import java.sql.SQLException
import java.sql.Statement
import java.util.Arrays
import java.util.HashSet
import java.util.Set

import com.branegy.dbmaster.connection.*
import com.branegy.dbmaster.model.*
import com.branegy.dbmaster.model.ForeignKey.ReferentialAction
import com.branegy.dbmaster.model.ForeignKey.ColumnMapping
import com.branegy.dbmaster.model.Index.IndexColumn
import com.branegy.dbmaster.model.Index.IndexType
import com.branegy.dbmaster.model.Parameter.ParamType
import com.branegy.dbmaster.util.NameMap
import com.branegy.service.core.exception.*
import org.apache.commons.io.IOUtils
import java.util.AbstractMap.SimpleEntry

/*
 * SQL 2000: User has access to all definitions of the database 
 * SQL 2005: user "GRANT VIEW DEFINITION TO [<user>]" for database
 */

public class SQLServer2008Dialect extends SQLServerDialect {
    final Set<String> excludeProcedureSet = [
        "dbo.sp_alterdiagram",
        "dbo.sp_creatediagram",
        "dbo.sp_dropdiagram",
        "dbo.sp_helpdiagramdefinition",
        "dbo.sp_helpdiagrams",
        "dbo.sp_renamediagram",
        "dbo.sp_upgraddiagrams"] as Set;
    final Set<String> excludeFunctionSet = ["dbo.fn_diagramobjects"] as Set;
    final Set<String> excludeTableSet = ["dbo.sysdiagrams"] as Set;

    def rsToString = { data -> 
        if (data==null) { 
            return null
        } else if (data instanceof java.sql.Clob) {
            Reader reader = data.getCharacterStream();
            return IOUtils.toString(reader);
        } else {
          return data.toString()
        }
    }

    public SQLServer2008Dialect(JdbcConnector cp,Connection connection, String dialectName, String dialectVersion)
            throws SQLException {
        super(cp,connection, dialectName, dialectVersion);
        typesWithPrecesion = new HashSet<String>( ["FLOAT", "REAL"] );
        typesWithScale = new HashSet ( [ "NUMERIC", "DECIMAL" ] );
        typesWithSize = new HashSet   ([ "BINARY", "CHAR", "DATETIME2", "DATETIMEOFFSET", 
                                         "DECIMAL", "NCHAR", "NUMERIC", "NVARCHAR", 
                                         "TIME", "VARBINARY", "VARCHAR" ]) 
    }

    @Override
    protected NameMap<Table> getTables(Connection conn, RevEngineeringOptions options) throws Exception {
        String sql =
                """ select s.name as schemaName,
                           t.name as tableName,
                           cast(p.value as varchar(2000)) as ${Table.DESCRIPTION},
                           t.create_date, 
                           t.modify_date
                      from sys.tables as t 
                      inner join sys.schemas as s on t.schema_id=s.schema_id
                      left join sys.extended_properties as p 
                             on t.object_id = p.major_id and p.name='MS_Description' and p.minor_id=0
                      where""";
        return this.getTables(conn, options, sql,"order by schemaName,tableName","s.name", "t.name",
            new SimpleEntry<>(Table.DESCRIPTION, String.class)
            //new ExtraCustomField(Table.ROWS, Integer.class),
        );
    }

    @Override
    protected void populateTableColumns(Connection conn, RevEngineeringOptions options, 
                                        NameMap<Table> tables) throws SQLException {

        String sql =
              """select sch.name as schema_name,
                        tab.name as table_name,
                        col.name as column_name,  
                        systyp.name as type_name, 
                        typ.is_user_defined, 
                        typ.name as user_type, 
                        col.max_length, 
                        col.precision, 
                        col.scale, 
                        col.collation_name, 
                        col.is_nullable as nullable, 
                        col.is_identity,
                        col.is_computed,
                        cc.definition as computedExpression, 
                        cc.is_persisted,
                        cc.uses_database_collation,
                        col.is_rowguidcol,
                        def.definition as defaultValue,
                        cast(ep.value as varchar(7500)) as comment
            from sys.columns as col  
            inner join sys.tables as tab on col.object_id=tab.object_id  
            inner join sys.schemas as sch on tab.schema_id = sch.schema_id  
            inner join sys.types as typ on col.user_type_id= typ.user_type_id  
            inner join sys.types as systyp on col.system_type_id  = systyp.system_type_id  
            and systyp.is_user_defined=0 and systyp.user_type_id = systyp.system_type_id  
            left join sys.default_constraints as def on col.default_object_id = def.object_id 
            left join sys.extended_properties as ep on ep.major_id=tab.object_id 
                      and ep.minor_id=col.column_id 
                      and ep.class='1' and ep.class_desc='OBJECT_OR_COLUMN' and ep.name='MS_Description' 
            left join sys.computed_columns as cc 
                      on tab.object_id=cc.object_id and col.column_id = cc.column_id
            where""";

        this.populateTableColumns(conn,options,tables,sql,"order by schema_name, table_name, col.column_id",
            "sch.name", "tab.name");
    }

    @Override
    protected NameMap<View> getViews(Connection conn, RevEngineeringOptions options) throws Exception {
        String sql =
                """select 
                        s.name as schemaName,
                        v.name as viewName,  
                        v.create_date,
                        v.modify_date,
                        c.text as viewText,
                        cast(p.value as varchar(2000)) as viewDescription,
                        c.colid
                   from sys.views as v  
                   inner join sys.schemas as s on v.schema_id=s.schema_id 
                   left join sys.syscomments as c on v.object_id=c.id 
                   left join sys.extended_properties as p   
                       on v.object_id = p.major_id and p.name='MS_Description' and p.minor_id=0
                   where v.type='V' AND""";

        return this.getViews(conn, options,sql,"order by s.name, v.name, colid","s.name", "v.name");
    }

    
    
    @Override
    protected NameMap<Procedure> getProcedures(Connection conn, RevEngineeringOptions options) throws Exception {
        String sql =
                """select 
                    s.name as schemaName,  
                    pr.name as procedureName, 
                    pr.create_date, 
                    pr.modify_date, 
                    c.text as procedureText,
                    cast(p.value as varchar(2000)) as procedureDescription,
                    c.colid
                   from sys.procedures as pr 
                   inner join sys.schemas as s on pr.schema_id=s.schema_id 
                   left join sys.extended_properties as p 
                      on pr.object_id = p.major_id and p.name='MS_Description' and p.minor_id=0 
                   inner join sys.syscomments as c on pr.object_id=c.id
                   where""";
                   
       return this.getProcedures(conn, options, sql, "order by s.name, pr.name,colid", "s.name", "pr.name",
           "N'P',N'PC',N'X',N'RF'");
    }

    protected void populateParameters(Connection conn, RevEngineeringOptions options,
                                      NameMap dbObjects, String typeFilter) throws SQLException {
        String sql =
                """   select sch.name as schema_name,
                             obj.name as object_name,
                             col.name as column_name,
                             systyp.name as type_name,
                             col.max_length, 
                             col.precision, 
                             col.scale, 
                             col.is_xml_document, 
                             col.is_readonly, 
                             col.is_output,
                             cast(col.default_value as varchar(7500)) as defaultValue,
                             cast(ep.value as varchar(7500)) as comment
                   from sys.parameters as col 
                   inner join sys.objects as obj on obj.object_id=col.object_id 
                   inner join sys.schemas as sch on obj.schema_id = sch.schema_id 
                   inner join sys.types as typ on col.user_type_id= typ.user_type_id 
                   inner join sys.types as systyp on col.system_type_id  = systyp.system_type_id 
                   and systyp.is_user_defined=0 and systyp.user_type_id = systyp.system_type_id 
                   left join sys.extended_properties as ep 
                            on ep.major_id=obj.object_id 
                           and ep.minor_id=col.parameter_id 
                           and ep.class='1'
                           and ep.class_desc='OBJECT_OR_COLUMN' 
                           and ep.name='MS_Description'
                   where (len(col.name)>0 or col.is_output=0) and obj.type in ("""+typeFilter+") AND"

        this.populateParameters(conn,options,dbObjects,typeFilter, sql, null, "sch.name","col.name");
    }

    @Override
    protected void setTableColumnComments(Connection conn, String database, 
            NameMap<Table> tables) throws SQLException {
        // NOP
    }
    
    @Override
    protected void setViewColumnComments(Connection conn, String database, 
            NameMap<View> tables) throws SQLException {
        // NOP
    }

}